import json
import locale
import os
import tempfile
import webbrowser
from importlib import resources as pkg_resources
from os.path import expanduser, dirname, realpath, join, exists

import requests

from . import resources
from .. import instruments

_CHART_JS_URL = 'https://cdnjs.cloudflare.com/ajax/libs/Chart.js/4.2.1/chart.umd.min.js'
_CACHE_DIR = expanduser(join(dirname(realpath(__file__)), 'cache'))
_CHART_JS_CACHE_FILE = expanduser(join(_CACHE_DIR, 'chart.js'))
_MAIN_JS_FILE = expanduser(join(dirname(realpath(__file__)), 'resources/main.js'))
_MAIN_CSS_FILE = expanduser(join(dirname(realpath(__file__)), 'resources/main.css'))
_DEFAULT_LOCALE = locale.getlocale()[0].replace('_', '-')
try:
    locale.setlocale(locale.LC_ALL, '')
    _DEFAULT_CURRENCY = locale.localeconv().get('int_curr_symbol').strip()
except:
    _DEFAULT_CURRENCY = 'USD'


def generate(title: str,
             value_by_ticker: dict[str, float],
             currency: str = _DEFAULT_CURRENCY,
             user_locale: str = _DEFAULT_LOCALE):
    data_by_ticker = instruments.get_data(list(value_by_ticker.keys()))
    data = []
    for key, value in data_by_ticker.items():
        value['quantity'] = value_by_ticker[key]
        data.append(value)
    _generate_report(title, currency, user_locale, data)


def _generate_report(title: str, currency: str, user_locale: str, data: list[dict]):
    _ensure_chart_js_downloaded()
    report = pkg_resources.read_text(resources, 'report_template.html') \
        .replace('%DATA%', json.dumps(data)) \
        .replace('%CHART_JS%', _CHART_JS_CACHE_FILE) \
        .replace('%MAIN_JS%', _MAIN_JS_FILE) \
        .replace('%MAIN_CSS%', _MAIN_CSS_FILE) \
        .replace('%CURRENCY%', currency) \
        .replace('%LOCALE%', user_locale) \
        .replace('%TITLE%', title)
    with tempfile.NamedTemporaryFile(mode='w', delete=False) as file:
        file.write(report)
        webbrowser.open('file://' + file.name)


def _ensure_chart_js_downloaded():
    if exists(_CHART_JS_CACHE_FILE):
        return
    os.makedirs(name=_CACHE_DIR, mode=0o755, exist_ok=True)
    with open(_CHART_JS_CACHE_FILE, 'w') as file:
        file.write(requests.get(_CHART_JS_URL).text)



import json
import multiprocessing
import re
import sys
import time
from dataclasses import asdict

import countrynames
import pycountry
import requests

from ...cache import cache, CACHE_EXPIRATION
from ..model import InstrumentData, InstrumentDataSource


class FundsDataSource(InstrumentDataSource):
    def get(self, instruments: list[str]) -> dict[str, dict]:
        pool = multiprocessing.Pool(len(instruments))
        results = pool.map(_get_result, [instrument for instrument in instruments])
        pool.close()
        return_values = {}
        for index, instrument in enumerate(instruments):
            if results[index] is not None:
                return_values[instrument] = results[index]
        return return_values


def _get_result(instrument: str) -> dict | None:
    try:
        if instrument.startswith("FX"):
            return _finex(instrument)
        elif instrument.startswith("T"):
            return _tinkoff(instrument)
        else:
            return None
    except _InstrumentMissingException:
        return None


@cache.memoize(expire=CACHE_EXPIRATION)
def _finex(instrument: str) -> dict:
    url = "https://finex-etf.ru/products/" + instrument
    print("Sending request GET " + url)
    start = time.time()
    r = requests.get(url)
    print("Got response for " + url + " in " + str(time.time() - start) + " seconds")
    if r.status_code == 404:
        raise _InstrumentMissingException
    group = re.search('<script id="__NEXT_DATA__" type="application/json">([^<]*)</script>', r.text).group(1)
    data = json.loads(group)
    try:
        response_data = data['props']['pageProps']['initialState']['fondDetail']['responseData']
    except IndexError:
        raise _InstrumentMissingException
    country_share = response_data['share'].get('countryShare')
    if country_share is None or country_share == {}:
        try:
            country_share = {response_data['name'].split("/")[1].strip(): 1}
        except:
            country_share = {}
    return asdict(InstrumentData(
        instrument=instrument,
        countries=_map_keys(country_share, _country_name_to_english),
        industries=response_data['share'].get('otherShare'),
        fee=response_data['commission'],
        currencies={
            str(response_data['currencyNav']): 19
        },
        classes={
            str(response_data['classActive']): 1
        }
    ))


@cache.memoize(expire=CACHE_EXPIRATION)
def _tinkoff(instrument: str) -> dict:
    url = "https://www.tinkoff.ru/invest/etfs/" + instrument
    print("Sending request GET " + url)
    start = time.time()
    r = requests.get(url)
    print("Got response for " + url + " in " + str(time.time() - start) + " seconds")
    r.encoding = 'UTF-8'
    group = re.search("<script id=\"__REACT_QUERY_STATE__invest\" type=\"application/json\">(.*?)</script>",
                      r.text).group(1) \
        .replace('\\\\"', '')
    data = json.loads(group)
    try:
        response_data = data['queries'][0]['state']['data']['detail']
    except IndexError:
        raise _InstrumentMissingException
    return asdict(InstrumentData(
        instrument=instrument,
        countries=_map_keys(_tinkoff_chart_to_shares(response_data, 'countries'), _country_name_to_english),
        industries=_tinkoff_chart_to_shares(response_data, 'sectors'),
        fee=response_data['expense']['total'],
        currencies={
            str(response_data['currency']): 1
        },
        classes=_tinkoff_chart_to_shares(response_data, 'types')
    ))


def _tinkoff_chart_to_shares(response_data: dict, key: str) -> dict[str, float]:
    return {x['name']: round(x['relativeValue'] / 100, 8) for x in
            next(filter(lambda obj: obj['type'] == key, response_data['pies']['charts']))['items']}


class _InstrumentMissingException(Exception):
    pass


def _map_keys(dictionary: dict[str, any], mapping_function) -> dict[str, any]:
    return {mapping_function(k): v for k, v in dictionary.items()}


def _country_name_to_english(name: str) -> str:
    try:
        return pycountry.countries.get(alpha_2=countrynames.to_code(name)).name
    except LookupError:
        print('Unexpected country name: "' + name + '"', file=sys.stderr)
        return name


funds = FundsDataSource().get



import sys
import time
from dataclasses import asdict

import pycountry
import yfinance

from ...cache import cache, CACHE_EXPIRATION
from ..model import InstrumentData, InstrumentDataSource

_DEFAULT_EXCHANGE = "ME"  # todo parameterize it in some other way


class SecurityDataSource(InstrumentDataSource):
    def get(self, instruments: list[str]) -> dict[str, dict]:
        result = {}
        for instrument in instruments:
            try:
                result[instrument] = _yahoo(instrument,
                    instrument if instrument.__contains__(".") else instrument + "." + _DEFAULT_EXCHANGE)
            except _InstrumentMissingException:
                print('No data for ticker "' + instrument + '", allocation report will not reflect it', file=sys.stderr)
                continue
        return result


@cache.memoize(expire=CACHE_EXPIRATION)
def _yahoo(instrument: str, instrument_with_exchange: str) -> dict:
    print("Sending request to Yahoo Finance for " + instrument)
    start = time.time()
    info = yfinance.Ticker(instrument_with_exchange).get_info()
    print("Got response in " + str(time.time() - start) + " seconds")
    if info.get('quoteType') is None:
        raise _InstrumentMissingException
    info_keys = info.keys()
    return asdict(InstrumentData(
        instrument=instrument,
        countries={
            pycountry.countries.get(alpha_2='RU').name: 1  # todo it must not be always RU
        },
        industries=None if 'sector' not in info_keys else {
            info['sector']: 1
        },
        fee=0,
        currencies=None if 'financialCurrency' not in info_keys else {
            info['financialCurrency']: 1
        },
        classes=None if 'quoteType' not in info_keys else {
            info['quoteType']: 1
        }
    ))


class _InstrumentMissingException(Exception):
    pass


securities = SecurityDataSource().get



from dataclasses import asdict

import pycountry

from ..model import InstrumentData, InstrumentDataSource


class CurrencyDataSource(InstrumentDataSource):
    def get(self, instruments: list[str]) -> dict[str, dict]:
        result = {}
        for currency_code in instruments:
            currency = pycountry.currencies.get(alpha_3=currency_code)
            if currency is None:
                continue
            result[currency_code] = asdict(InstrumentData(
                instrument=currency_code,
                countries={
                    self._to_country(currency.alpha_3): 1
                },
                industries={},
                fee=0,
                currencies={
                    currency.alpha_3: 1
                },
                classes={
                    'Cash': 1
                }
            ))
        return result

    @staticmethod
    def _to_country(currency: str) -> str:
        return 'Europe' if currency == 'EUR' else pycountry.countries.get(alpha_2=currency[0:2]).name


currencies = CurrencyDataSource().get



from .sources import currencies, funds, securities, custom


def get_data(tickers: list[str]) -> dict[str, dict]:
    info_by_currency = currencies.currencies(tickers)
    _currencies = info_by_currency.keys()
    not_currencies = list(filter(lambda ticker: ticker not in _currencies, tickers))
    info_by_fund = funds.funds(not_currencies)
    _funds = info_by_fund.keys()
    other = list(filter(lambda ticker: ticker not in _funds, not_currencies))
    _securities = securities.securities(other)
    result = info_by_currency | info_by_fund | _securities
    custom_info = custom.custom(tickers)
    for key, value in result.items():
        if key in custom_info.keys():
            value.update(custom_info[key])
    return result



import argparse
import json
import os
import sys

from . import instruments, gnucash, report

_DEFAULT_REPORT_NAME_ENV_VAR = 'PORTFOLIO_ALLOCATION_GNUCASH_REPORT'


def main():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(title='commands', help='Commands description', required=True)

    asset_data = subparsers.add_parser('tickerdata',
                                       help='Shows currency, fee, allocation data for provided list of tickers')
    asset_data.set_defaults(cmd='tickerdata')
    asset_data.add_argument('tickers', metavar='ticker', type=str, nargs='+',
                            help='a ticker of an asset to get info for')

    asset_allocation_gnucash = subparsers.add_parser(
        'gnucash',
        help='Generates allocation report based on GnuCash\'s Security Piechart and allocation data of its components')
    asset_allocation_gnucash.set_defaults(cmd='gnucash')
    asset_allocation_gnucash.add_argument(
        "-r", "--report-names",
        help="Name of report which contains securities allocation. Default: Securities",
        nargs='+', type=str,
        default=[os.environ.get(_DEFAULT_REPORT_NAME_ENV_VAR, 'Securities')])

    default_datafile = gnucash.get_latest_file()
    if default_datafile is None:
        asset_allocation_gnucash.add_argument("-f", "--datafile", required=True, type=str,
                                              help="GnuCash datafile (.gnucash)")
    else:
        asset_allocation_gnucash.add_argument(
            "-f", "--datafile",
            help="GnuCash datafile (.gnucash). Default: " + default_datafile,
            nargs='?', const=1, type=str,
            default=default_datafile)

    args = parser.parse_args()

    if args.cmd == 'tickerdata':
        print(json.dumps(instruments.get_data(args.tickers), indent=2, ensure_ascii=False))
    elif args.cmd == 'gnucash':
        for report_name in args.report_names:
            try:
                parsed_gnucash_report = gnucash.get_value_by_instrument(report_name=report_name, datafile=args.datafile)
                report.generate(parsed_gnucash_report.title,
                                parsed_gnucash_report.value_by_instrument,
                                parsed_gnucash_report.currency)
            except gnucash.ParseException as e:
                print(e, file=sys.stderr)


if __name__ == '__main__':
    main()



import re
import subprocess
import sys
from dataclasses import dataclass


@dataclass
class ParsedGnuCashReport:
    title: str
    currency: str
    value_by_instrument: dict[str, float]


class ParseException(Exception):
    pass


class CannotRunGnuCash(Exception):
    pass


def get_value_by_instrument(report_name: str, datafile: str) -> ParsedGnuCashReport:
    proc = _try_run_gnucash_cli(['--report', 'run', '--name=' + report_name, datafile])
    return parse_value_by_instrument(proc.stdout.read().decode('utf-8'))


def _try_run_gnucash_cli(params: [str]):
    cmd_list = [['gnucash-cli'], ['flatpak', 'run', '--command=gnucash-cli', 'org.gnucash.GnuCash']]
    for cmd in cmd_list:
        try:
            cmd.extend(params)
            return subprocess.Popen(cmd, stdout=subprocess.PIPE)
        except:
            continue
    raise CannotRunGnuCash()


def parse_value_by_instrument(html_report: str) -> ParsedGnuCashReport:
    labels_match = re.search('"labels"\\s*:\\s*\\[\\s*"(.*)"\\s*],', html_report)
    if labels_match is None:
        raise ParseException("Could not find chart's labels in GnuCash's report")
    labels = re.split('"\\s*,\\s*"', labels_match.group(1))
    instruments = [re.split('\\s*-\\s*', label)[0] for label in labels]
    volumes = list(
        map(float, re.split('\\s*,\\s*', re.search('"data"\\s*:\\s*\\[\\s*(.*)\\s*],', html_report).group(1))))
    currency = re.search('var\\s+curriso\\s*=\\s*"(.*)"\\s*;', html_report).group(1)
    title = re.search('"text"\\s*:\\s*\\[\\s*"(.*)"\\s*,\\s*"(.*)"\\s*]', html_report).group(1)
    return ParsedGnuCashReport(title, currency, dict(zip(instruments, volumes)))


def get_latest_file():
    # from https://github.com/sdementen/gnucash-utilities/blob/develop/piecash_utilities/config.py
    if sys.platform.startswith("linux"):
        import subprocess
        try:
            output_dconf = subprocess.check_output(["dconf", "dump", "/org/gnucash/GnuCash/history/"]).decode()
            from configparser import ConfigParser
            conf = ConfigParser()
            conf.read_string(output_dconf)
            return conf["/"]["file0"][1:-1]
        except:
            return None
    else:
        return None



from dataclasses import dataclass


@dataclass
class InstrumentData:
    instrument: str
    countries: dict[str, float]
    industries: dict[str, float]
    fee: float
    currencies: dict[str, float]
    classes: dict[str, float]


class InstrumentDataSource:
    def get(self, instruments: list[str]) -> dict[str, dict]:
        pass



from os.path import expanduser, join, dirname, realpath

from diskcache import Cache

_CACHE_DIR = expanduser(join(dirname(realpath(__file__)), 'cache'))

cache = Cache(_CACHE_DIR)

CACHE_EXPIRATION = 60 * 60 * 24 * 30 # 30 days



import json
import os

from ..model import InstrumentDataSource

_SECURITIES_CUSTOM_JSON = os.path.join(
    os.environ.get('APPDATA') or
    os.environ.get('XDG_CONFIG_HOME') or
    os.path.join(os.environ['HOME'], '.config'),
    "portfolio-allocation",
    "securities-custom.json"
)


class CustomDataSource(InstrumentDataSource):
    def get(self, instruments: list[str]) -> dict[str, dict]:
        if not os.path.exists(_SECURITIES_CUSTOM_JSON):
            print("No custom config \"" + _SECURITIES_CUSTOM_JSON + "\", ignoring it")
            return {}
        print("Reading custom config \"" + _SECURITIES_CUSTOM_JSON + "\"")
        f = open(_SECURITIES_CUSTOM_JSON, "r")
        configs = json.loads(f.read())
        result = {}
        for config in configs:
            instrument_in_config = config['instrument']
            if instrument_in_config == '*':
                for i in instruments:
                    result[i] = config.copy()
                    result[i]['instrument'] = i
            elif instrument_in_config.endswith('*'):
                for i in instruments:
                    if i.startswith(instrument_in_config[:-1]):
                        result[i] = config.copy()
                        result[i]['instrument'] = i
            elif instrument_in_config in instruments:
                result[instrument_in_config] = config.copy()
        return result


custom = CustomDataSource().get
